#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2021 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Awaitable, Callable
import functools
import json
import typing

import hiyori

from . import _methods
from ._collections import AttrDict, Json
from ._exceptions import FailedRequest, MalformedResponse, TelemeException
from ._version import __version__

__all__ = [
    "Api",
    "AttrDict",
    "TelemeException",
    "MalformedResponse",
    "FailedRequest",
    "__version__",
]

_USER_AGENT = "teleme/{} hiyori/{}".format(__version__, hiyori.__version__)

_DEFAULT_BASE_URL = "https://api.telegram.org/bot{token}/{method}"


class Api:
    def __init__(
        self,
        token: str,
        *,
        base_url: str = _DEFAULT_BASE_URL,
        user_agent: str = _USER_AGENT,
        timeout: int = 61,
    ) -> None:
        self._token = token
        self._base_url = base_url
        self._user_agent = user_agent

        self._client = hiyori.HttpClient(timeout=timeout)

    def _make_request_url(self, method_name: str) -> str:
        try:
            url_name = _methods.get_url_name(method_name)

        except ValueError as e:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute "
                f"'{method_name}'"
            ) from e

        return self._base_url.format(token=self._token, method=url_name)

    async def _send_anything(
        self, __request_url: str, __method_name: str, **kwargs: Any
    ) -> Json:
        """
        If you want to determine which method will be invoked at runtime,
        please use :code:`getattr(Api(), "method_name")` instead of this
        function.
        """

        # Telegram will cut off requests longer than 60 seconds,
        # but it's better to wait the server close the connection first.
        headers = {
            "User-Agent": self._user_agent,
            "Accept": "application/json",
        }

        r = await self._client.post(
            __request_url, headers=headers, json=kwargs
        )

        try:
            j = json.loads(r.body.to_str(), object_pairs_hook=AttrDict)

        except json.decoder.JSONDecodeError as e:
            raise MalformedResponse(
                status_code=r.status_code,
                method=__method_name,
                body=r.body.to_str(),
                sent_kwargs=kwargs,
            ) from e

        except UnicodeDecodeError as e:
            raise MalformedResponse(
                status_code=r.status_code,
                method=__method_name,
                body=r.body,
                sent_kwargs=kwargs,
            ) from e

        if not isinstance(j, AttrDict) or "ok" not in j.keys():
            raise MalformedResponse(
                status_code=r.status_code,
                method=__method_name,
                body=j,
                sent_kwargs=kwargs,
            )

        if not j.ok or r.status_code != 200:
            raise FailedRequest(
                status_code=r.status_code,
                method=__method_name,
                body=j,
                sent_kwargs=kwargs,
            )

        return typing.cast(Json, j.result)

    def __getattr__(self, name: str) -> Callable[..., Awaitable[Json]]:
        return functools.partial(
            self._send_anything, self._make_request_url(name), name
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2021 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Dict
import re
import threading

_GLOBAL_LOCK = threading.Lock()

_NAME_RE = re.compile(r"^[a-z]([a-z\_]+)?$")

_CACHED_METHODS: Dict[str, str] = {}


def get_url_name(attr_name: str) -> str:
    """
    Translate pep8 compliant names to telegram api method names.
    """
    with _GLOBAL_LOCK:
        if attr_name not in _CACHED_METHODS.keys():
            if _NAME_RE.fullmatch(attr_name) is None:
                raise ValueError(
                    f"Unacceptable API Method Name: `{attr_name}`, "
                    f'r"{_NAME_RE.pattern}" is expected.'
                )

            _CACHED_METHODS[attr_name] = attr_name.replace("_", "")

        return _CACHED_METHODS[attr_name]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2021 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Dict, Union

from ._collections import AttrDict, Json


class TelemeException(Exception):
    def __init__(
        self,
        *,
        status_code: int,
        method: str,
        body: Union[AttrDict[Json], str, bytes],
        sent_kwargs: Dict[str, Any],
    ) -> None:
        self.method = method
        self.sent_kwargs = sent_kwargs

        self.status_code = status_code
        self.body = body

        super().__init__(
            f"Request: {self.method}({repr(self.sent_kwargs)})\n"
            f"Status Code: {self.status_code}\n"
            f"Response Body: \n{repr(self.body)}"
        )


class FailedRequest(TelemeException):
    pass


class MalformedResponse(TelemeException):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['teleme']

package_data = \
{'': ['*']}

install_requires = \
['hiyori>=0.2,<0.3']

extras_require = \
{':python_version <= "3.7"': ['importlib-metadata>=4.0.1,<5.0.0']}

setup_kwargs = {
    'name': 'teleme',
    'version': '0.2.0',
    'description': 'An async, super simple Telegram Bot framework.',
    'long_description': 'Teleme\n======\nTeleme(pronounced `tell-me`)  is an async, super simple Telegram Bot framework.\n\nUnder Development\n-----------------\nWhilst the code has been used for years privately, since I just released the code,\nthere may have been some bugs that have left undiscovered. I have decided to set\nthe status of this project as Beta.\n\nLicense\n-------\n\n    Copyright 2020 Kaede Hoshikawa\n\nLicensed under the Apache License, Version 2.0 (the "License");\nyou may not use this file except in compliance with the License.\nYou may obtain a copy of the License at\n\n    http://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software\ndistributed under the License is distributed on an "AS IS" BASIS,\nWITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\nSee the License for the specific language governing permissions and\nlimitations under the License.\n',
    'author': 'Kaede Hoshikawa',
    'author_email': 'futursolo@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/futursolo/teleme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2021 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Dict, Sequence, TypeVar, Union
import typing

from typing_extensions import Protocol


class _JsonList(Protocol):
    def __getitem__(self, idx: int) -> "Json":
        ...

    # hack to enforce an actual list
    def sort(self) -> None:
        ...


class _JsonDict(Protocol):
    def __getitem__(self, key: str) -> "Json":
        ...

    # hack to enforce an actual dict
    @staticmethod
    @typing.overload
    def fromkeys(seq: Sequence[Any]) -> Dict[Any, Any]:
        ...

    @staticmethod  # noqa: F811
    @typing.overload
    def fromkeys(seq: Sequence[Any], value: Any) -> Dict[Any, Any]:
        ...


Json = Union[str, int, float, bool, None, _JsonList, _JsonDict]

_TJson = TypeVar("_TJson", bound=Json)


class AttrDict(Dict[str, _TJson]):
    def __getitem__(self, name_or_index: str) -> _TJson:
        try:
            value = dict.__getitem__(self, name_or_index)

        except KeyError:
            if name_or_index != "_from":
                raise

            value = self["from"]

        return value

    def __getattr__(self, name_or_index: str) -> _TJson:
        try:
            return self[name_or_index]

        except KeyError as e:
            raise AttributeError from e
